from __future__ import annotations

import subprocess
import time
from datetime import datetime, timedelta
from functools import cached_property, wraps
from typing import Any, Callable, Optional

from mower import __system__
from mower.utils import config
from mower.utils import typealias as tp
from mower.utils.csleep import MowerExit, csleep
from mower.utils.device.exception import ADBError, GameError, SimulatorError
from mower.utils.device.method.adb import ADB
from mower.utils.device.method.droidcast import DroidCast
from mower.utils.device.method.mumu_ipc import (
    MuMu12IPC,
)
from mower.utils.device.method.scrcpy import Scrcpy
from mower.utils.device.simulator import restart_simulator
from mower.utils.image import bytes2img
from mower.utils.log import logger, save_screenshot


def retry_adb(func):
    @wraps(func)
    def retry_wrapper(self, *args, **kwargs):
        for attempt in range(2):
            try:
                return func(self, *args, **kwargs)
            except MowerExit:
                raise
            except ADBError as e:
                logger.warning(f"ADB error, retrying... ({attempt + 1}/3)")
                logger.exception(e)
                self.adb.adb_reconnect()
            except Exception as e:
                raise Exception(e)
        # After adb retries fail, escalate to simulator error
        raise SimulatorError("ADB failed after 2 attempts.")

    return retry_wrapper


def retry(func):
    @wraps(func)
    def retry_wrapper(self, *args, **kwargs):
        for attempt in range(2):
            try:
                return func(self, *args, **kwargs)
            except MowerExit:
                raise
            except GameError as e:
                logger.warning(f"Game error, retrying... ({attempt + 1}/2)")
                logger.exception(e)
                self.check_current_focus()
            except SimulatorError as e:
                logger.warning(f"Simulator error, retrying... ({attempt + 1}/2)")
                logger.exception(e)
                restart_simulator()
            except Exception as e:
                logger.exception(e)
                restart_simulator()
        raise MowerExit

    return retry_wrapper


class Device:
    solver_name = "设备"

    @cached_property
    def adb(self) -> ADB:
        return ADB()

    @cached_property
    def droidcast(self) -> DroidCast:
        return DroidCast(self.adb)

    @cached_property
    def scrcpy(self) -> Scrcpy:
        return Scrcpy(self.adb)

    @cached_property
    def mumu_ipc(self) -> MuMu12IPC:
        return MuMu12IPC()

    @retry
    def run(self, cmd: str) -> Optional[bytes]:
        return self.adb.adb_shell(cmd)

    def launch(self) -> None:
        """launch the application"""
        logger.info("明日方舟，启动！")

        tap = config.conf.tap_to_launch_game.enable
        x = config.conf.tap_to_launch_game.x
        y = config.conf.tap_to_launch_game.y

        if tap:
            self.run(f"input tap {x} {y}")
        else:
            self.run(
                f"monkey --pct-syskeys 0 -p {config.conf.APPNAME} -c android.intent.category.LAUNCHER 1"
            )

    def exit(self) -> None:
        """exit the application"""
        logger.info("退出游戏")
        self.run(f"am force-stop {config.conf.APPNAME}")

    def home(self) -> None:
        """back to home"""
        logger.info("返回模拟器主页")
        self.run(
            "am start -a android.intent.action.MAIN -c android.intent.category.HOME"
        )

    @retry
    @retry_adb
    def send_keyevent(self, keycode: int) -> None:
        """send a key event"""
        logger.debug(keycode)
        self.scrcpy.control.send_keyevent(keycode)

    def send_text(self, text: str) -> None:
        """send a text"""
        logger.debug(repr(text))
        text = text.replace('"', '\\"')
        command = f'input text "{text}"'
        self.run(command)

    def diy_capture_display():
        command = config.conf.custom_screenshot.command
        data = subprocess.check_output(
            command,
            shell=True,
            creationflags=subprocess.CREATE_NO_WINDOW if __system__ == "windows" else 0,
        )
        return bytes2img(data)

    @retry
    @retry_adb
    def screencap(self) -> bytes:
        start_time = datetime.now()
        min_time = config.screenshot_time + timedelta(
            milliseconds=config.conf.screenshot_interval
        )
        delta = (min_time - start_time).total_seconds()
        if delta > 0:
            time.sleep(delta)
            start_time = min_time

        if config.conf.screencap_strategy == "droidcast":
            img = self.droidcast.capture_display()
        elif config.conf.screencap_strategy == "mumuipc":
            img = self.mumu_ipc.capture_display()
        elif config.conf.screencap_strategy == "diy":
            img = self.diy_capture_display()
        else:
            img = self.adb.capture_display_with_gzip()

        stop_time = datetime.now()
        config.screenshot_time = stop_time
        interval = (stop_time - start_time).total_seconds() * 1000
        if config.screenshot_avg is None:
            config.screenshot_avg = interval
        else:
            config.screenshot_avg = config.screenshot_avg * 0.9 + interval * 0.1
        if config.screenshot_count >= 100:
            config.screenshot_count = 0
            logger.info(
                f"截图用时{interval:.0f}ms 平均用时{config.screenshot_avg:.0f}ms"
            )
        else:
            config.screenshot_count += 1

        save_screenshot(img)
        return img

    def current_focus(self) -> str:
        """detect current focus app"""
        command = "dumpsys window | grep mCurrentFocus"
        line = self.run(command)
        return line.strip()[:-1].split(" ")[-1]

    def display_frames(self) -> tuple[int, int, int]:
        """get display frames if in compatibility mode"""
        if not config.MNT_COMPATIBILITY_MODE:
            return None

        command = "dumpsys window | grep DisplayFrames"
        line = self.run(command)
        """ eg. DisplayFrames w=1920 h=1080 r=3 """
        res = line.strip().replace("=", " ").split(" ")
        return int(res[2]), int(res[4]), int(res[6])

    @retry
    @retry_adb
    def tap(self, point: tuple[int, int]) -> None:
        """tap"""
        logger.debug(point)
        self.scrcpy.tap(point[0], point[1])

    @retry
    @retry_adb
    def back(self):
        logger.debug("back")
        self.scrcpy.back()

    @retry
    @retry_adb
    def swipe(
        self, start: tuple[int, int], end: tuple[int, int], duration: int = 100
    ) -> None:
        """swipe"""
        logger.debug(f"{start=} {end=} {duration=}")
        self.scrcpy.swipe(start[0], start[1], end[0], end[1], duration / 1000)

    @retry
    @retry_adb
    def swipe_ext(
        self,
        points: list[tp.Coordinate],
        durations: list[int],
        update: bool = False,
        interval: float = 0,
        func: Callable[[tp.Image], Any] = lambda _: None,
    ):
        """复杂的滑动

        Args:
            points (list[tp.Coordinate]): 坐标列表
            durations (list[int]): 每两个坐标之间拖动的时长
            update (bool, optional): 抬手前截图. Defaults to False.
            interval (float, optional): 抬手后等待回弹动画的时间. Defaults to 0.
            func (Callable[[tp.Image], Any], optional): 处理抬手前截图的函数. Defaults to lambda _: None.

        Returns:
            Any: func的返回值
        """
        logger.debug(f"{points=} {durations=} {update=} {interval=} {func=}")
        return self.scrcpy.swipe_ext(points, durations, update, interval, func)

    def check_current_focus(self) -> bool:
        """check if the application is in the foreground"""
        update = False
        start_time = datetime.now()
        while True:
            try:
                focus = self.current_focus()
                if focus not in [
                    f"{config.conf.APPNAME}/{config.APP_ACTIVITY_NAME}",
                    "com.hypergryph.arknights.bilibili/com.gsc.welcome.WelcomeActivity",
                ]:
                    if (datetime.now() - start_time).total_seconds() > 25:
                        self.exit()  # 应用卡死
                        start_time = datetime.now()
                    self.launch()
                    update = True
                    csleep(1)
                    continue
                return update
            except MowerExit:
                raise
            except Exception as e:
                logger.exception(e)
                update = True

    def check_device_screen(self) -> bool:
        """检查分辨率和DPI

        Returns:
            设置是否正确
        """
        res_output = self.run("wm size").strip()
        dpi_output = self.run("wm density").strip()
        logger.debug(f"{res_output} {dpi_output}".replace("\n", " "))

        physical_res, _, override_res = res_output.partition("Override")
        physical_dpi, _, override_dpi = dpi_output.partition("Override")
        res = (override_res or physical_res).partition("size:")[2].strip()
        dpi = (override_dpi or physical_dpi).partition("density:")[2].strip()

        if res in ["1920x1080", "1080x1920"] and dpi == "280":
            return True
        logger.error("mower-ng仅支持1920x1080分辨率、280DPI，请修改模拟器设置")
        return False
