import ctypes
import os

import numpy as np

from mower.utils import config
from mower.utils.device.exception import GameError, SimulatorError
from mower.utils.log import logger


class NemuIpcIncompatible(Exception):
    pass


class ImageCaptureError(Exception):
    pass


class MuMu12IPC:
    def __init__(self):
        self.simulator_folder = config.conf.simulator.simulator_folder
        self.instanse_index = int(config.conf.simulator.index)
        self.connection = 0
        self.display_id = -1
        # 加载动态链接库
        dll_path = os.path.join(
            self.simulator_folder, "sdk", "external_renderer_ipc.dll"
        )
        try:
            self.external_renderer = ctypes.CDLL(dll_path)
        except OSError as e:
            logger.error(e)
            # OSError: [WinError 126] 找不到指定的模块。
            if not os.path.exists(dll_path):
                raise NemuIpcIncompatible(
                    f"ipc_dll={dll_path} does not exist, "
                    f"NemuIpc requires MuMu12 version >= 3.8.13, please check your version"
                )
            else:
                raise NemuIpcIncompatible(
                    f"ipc_dll={dll_path} exists, but cannot be loaded"
                )

        # 定义函数原型
        self.external_renderer.nemu_connect.argtypes = [ctypes.c_wchar_p, ctypes.c_int]
        self.external_renderer.nemu_connect.restype = ctypes.c_int

        self.external_renderer.nemu_disconnect.argtypes = [ctypes.c_int]
        self.external_renderer.nemu_disconnect.restype = None

        self.external_renderer.nemu_get_display_id.argtypes = [
            ctypes.c_int,
            ctypes.c_char_p,
            ctypes.c_int,
        ]
        self.external_renderer.nemu_get_display_id.restype = ctypes.c_int

        self.external_renderer.nemu_capture_display.argtypes = [
            ctypes.c_int,
            ctypes.c_uint,
            ctypes.c_int,
            ctypes.POINTER(ctypes.c_int),
            ctypes.POINTER(ctypes.c_int),
            ctypes.POINTER(ctypes.c_ubyte),
        ]
        self.external_renderer.nemu_capture_display.restype = ctypes.c_int

    def connect(self):
        "连接到 emulator"
        self.connection = self.external_renderer.nemu_connect(
            ctypes.c_wchar_p(os.path.dirname(self.simulator_folder)),
            self.instanse_index,
        )
        if self.connection == 0:
            raise SimulatorError(
                "连接模拟器失败，请检查模拟器文件夹是否正确，并且模拟器已经启动"
            )

    def get_display_id(self) -> int:
        pkg_name = config.conf.APPNAME.encode("utf-8")  # 替换为实际包名
        self.display_id = self.external_renderer.nemu_get_display_id(
            self.connection, pkg_name, 0
        )
        if self.display_id < 0:
            raise GameError("获取Display ID失败,请先打开游戏")
        logger.info(f"获取Display ID成功: {self.display_id}")

    def check_status(self):
        if self.connection == 0:
            self.connect()
        if self.display_id < 0:
            self.get_display_id()

    def capture_display(self) -> np.ndarray | None:
        self.check_status()
        pixels = (ctypes.c_ubyte * 8294400)()
        result = self.external_renderer.nemu_capture_display(
            self.connection,
            self.display_id,
            8294400,
            ctypes.byref(ctypes.c_int(1920)),
            ctypes.byref(ctypes.c_int(1080)),
            pixels,
        )
        if result != 0:
            self.connection = 0
            self.display_id = -1
            raise ImageCaptureError("捕获失败")

        image = np.frombuffer(pixels, dtype=np.uint8).reshape((1080, 1920, 4))[:, :, :3]
        image = np.flipud(image)  # 翻转
        return image
